#ifndef _EM
#define _EM

#include <ctime>
#include <stdexcept>
#include <string>

class Invalid_temperature_exception : std::exception { // '#include <stdexcept>'
    float _value; // In Celsius
public:
    Invalid_temperature_exception(float);
    float value() const;
    const char* what() const noexcept(true) final; // No exception can be thrown!
};

class Temperature {
public:

    enum class Temperature_unit : char /* 'char' -> Integral type is required! */ {
        Celsius = 'C', Fahrenheit = 'F', Kelvin = 'K'
    };
    static const float Min;
private:
    float _value; // In Celsius
    float _step;
public:
    float asCelsius() const;
    float asFahrenheit() const;
    float asKelvin() const;
    void increment() noexcept(true);
    void decrement() noexcept(false);
    int operator<(const Temperature&) const;
    int operator<=(const Temperature&) const;
    int operator>(const Temperature&) const;
    int operator>=(const Temperature&) const;
    Temperature();
    Temperature(float, Temperature_unit = Temperature_unit::Celsius) noexcept(false);
};

class Global_exception_management { // Design patten: 'Singleton', i.e., no instance required
public:
    static void my_unexpected();
    static void my_terminate();
};

class Invalid_temperature_switching_exception { // Caution: no inheritance here
private:
    Temperature::Temperature_unit _invalid_temperature_mode;
public:
    Invalid_temperature_switching_exception(Temperature::Temperature_unit);
    std::underlying_type<Temperature::Temperature_unit>::type invalid_temperature_mode() const;
};

class Programmable_thermostat {
private:

    struct Program {
        Temperature target_temperature;
        std::time_t time;

        /** For test only: */
        //        inline void f() {
        //            if (_Max < _Min);
        //        }
        /** End of test */
    };
public:
    static Temperature _Max;
    static Temperature _Min;
private:
protected:
    Program _program[8];
private:
    std::time_t _time;
    Temperature::Temperature_unit _temperature_mode;
    Temperature temperature() const;
    Temperature target_temperature;
public:
    Programmable_thermostat();
    void f_c() noexcept(false);
    bool temp_down();
    bool temp_up();
    void time_backward();
    void time_forward();
    std::string printable_time() const;
    std::string printable_temperature() const;
    std::string printable_target_temperature() const;
};

#endif

