#include <cstdlib>

#include <iostream>

#include "EM.h"

void callee() noexcept(false); // Equivalent declaration: 'void callee();'

void caller() noexcept(true) {
    ::callee(); // No compilation error...
    // throw "Something while 'noexcept(true)'"; // Compilation error...
};

template<typename T> T my_function() noexcept(sizeof (T) < 4);
// Instantiation of 'my_function' with 'T' = 'void'
decltype(::my_function<void>()) *p; // Compilation error because 'sizeof(void)'

int main(int argc, char** argv) {
    std::set_unexpected(&Global_exception_management::my_unexpected);
    std::set_terminate(&Global_exception_management::my_terminate);

    std::cout << "Absolute zero: " << Temperature::Min << std::endl;

    std::exception_ptr pe; // C++11
    try {
        // Instance of 'Invalid_temperature_exception' may potentially be raised (or "Illegal temperature unit"):
        Temperature t(0.F, Temperature::Temperature_unit::Kelvin);
        t.decrement(); // Value of 'float' *IS* raised...
    } catch (const char* e) { // Matches 'throw "Illegal temperature unit"'
        std::cerr << "const char*: " << e << std::endl;
    } catch (Invalid_temperature_exception& ite) { // Matches 'throw Invalid_temperature_exception(_value)'
        std::cerr << ite.what() << std::endl;
    }//    catch (float& e) { // Matches 'throw _value' 
    //        std::cerr << "float&: " << e << std::endl;
    //    }
    catch (std::bad_exception& be) { // 'Global_exception_management::my_unexpected' is called before
        std::cerr << "be: " << be.what() << std::endl;
    } catch (...) {
        std::cerr << "Default management" << std::endl;
        pe = std::current_exception(); // C++11
    }
    try {
        if (pe) std::rethrow_exception(pe);
    }
    // Great caution: missing 'throw _value':
    catch (const std::exception& e) {
        std::cout << "Caught exception \"" << e.what() << "\"\n";
    }

    try {
        Programmable_thermostat pt;
        std::cout << pt.printable_time().c_str() << std::endl;
        pt.time_forward();
        std::cout << pt.printable_time().c_str() << std::endl;
        std::cout << pt.printable_target_temperature().c_str() << std::endl;

        try {
            pt.f_c();
        } catch (Invalid_temperature_switching_exception& itse) {
            std::cerr << itse.invalid_temperature_mode();
            std::exit(1);
        }
        std::cout << pt.printable_target_temperature().c_str() << std::endl;
    } catch (...) {
        std::abort();
    }

    return 0;
}

